from __future__ import annotations

import dataclasses
import functools
from os import PathLike
from pathlib import Path
from string import capwords
from typing import Sequence

import dask.dataframe
import toolz
import pandas as pd


def isna(x):
    r = pd.isna(x)
    return any(r) if isinstance(x, Sequence) else r


@toolz.curry
def maybe(f, val):
    return f(val) if not isna(val) else val


@dataclasses.dataclass(frozen=True)
class ExportedData:

    root: Path

    @classmethod
    def from_path(cls, x: str | PathLike) -> ExportedData:
        return cls(Path(x).absolute())

    def drop_cached_data(self):
        cached_keys = set(self.__dict__) - {
            field.name for field in dataclasses.fields(self)
        }
        for k in cached_keys:
            del self.__dict__[k]

    # Aggregate views

    @functools.cached_property
    def daily_exercise(self) -> pd.DataFrame:
        return (
            self.exercise_log.drop(
                columns=[
                    "logId",
                    "activityName",
                    "activityTypeId",
                    "tcxLink",
                    "logType",
                    "hasGps",
                    "shouldFetchDetails",
                ]
            )
            .resample("D", on="startTime")
            .sum()
        )

    def _resample_log(
        self, log_name: str, resample_rule: str, set_index: bool = True
    ) -> pd.DataFrame:
        return getattr(self, log_name).resample(resample_rule).sum()

    @functools.cached_property
    def daily_steps(self) -> dask.dataframe.DataFrame:
        return self._resample_log("step_log", "D").compute()

    @functools.cached_property
    def hourly_steps(self) -> dask.dataframe.DataFrame:
        return self._resample_log("step_log", "60Min")

    @functools.cached_property
    def daily_calories(self) -> dask.dataframe.DataFrame:
        return self._resample_log("calorie_log", "D")

    @functools.cached_property
    def daily_sleep(self) -> dask.dataframe.DataFrame:
        return (
            self.sleep_log.drop(columns=["logId", "sleep_log_entry_id"])
            .resample("D")
            .first()
        )

    # Level 2 data
    # FitBit models that take Level 1 data as input

    @functools.cached_property
    def sleep_log(self) -> pd.DataFrame:
        log = self._transform_sleep_columns(
            dask.dataframe.read_json(
                self.root / "Sleep" / "sleep-*.json",
                lines=False,
                orient="records",
            ).compute()
            # For some reason some sleeps are included more than once
            .drop_duplicates(subset="logId")
        )
        scores = pd.read_csv(
            self.root / "Sleep" / "sleep_score.csv",
        )
        return (
            pd.merge(
                log,
                scores,
                left_on="logId",
                right_on="sleep_log_entry_id",
            )
            .drop(
                # Already stored in startTime
                columns="timestamp"
            )
            .set_index("startTime", drop=False)
        )

    @staticmethod
    def _transform_sleep_columns(df: pd.DataFrame) -> pd.DataFrame:
        # This function mutates its input.
        tc = toolz.curried

        # Flatten the sleep level summary data
        sleep_levels = df.pop("levels")
        levels = ["wake", "light", "rem", "deep"]
        for level in levels:
            df[level + "Count"] = sleep_levels.map(
                tc.get_in(["summary", level, "count"])
            )
            df[level + "Minutes"] = sleep_levels.map(
                tc.get_in(["summary", level, "minutes"])
            )
        # keep the detailed sleep level data for time-series analysis
        df["detail"] = sleep_levels.map(tc.get_in(["data"]))

        df["startTime"] = pd.to_datetime(df["startTime"])
        df["endTime"] = pd.to_datetime(df["endTime"])
        del df["dateOfSleep"]

        return df

    @functools.cached_property
    def exercise_log(self):
        return toolz.thread_last(
            Path.glob(Path("data") / "Physical Activity", "exercise-*.json"),
            (map, toolz.partial(pd.read_json, lines=False)),
            pd.concat,
            self._transform_exercise_columns,
        )

    @staticmethod
    def _transform_exercise_columns(df: pd.DataFrame) -> pd.DataFrame:
        # This function mutates its input.
        tc = toolz.curried

        activityLevel = df.pop("activityLevel")
        levels = ["sedentary", "lightly", "fairly", "very"]
        for i, level in enumerate(levels):
            df[f"minutes{capwords(level)}Active"] = activityLevel.map(
                maybe(tc.get_in([i, "minutes"]))
            )

        heartRateZones = df.pop("heartRateZones")
        zones = ["Out of Range", "Fat Burn", "Cardio", "Peak"]
        for i, zone in enumerate(zones):
            df[f"Minutes in {zone} Zone"] = heartRateZones.map(
                maybe(tc.get_in([i, "minutes"]))
            )

        df.drop(columns="manualValuesSpecified", inplace=True)
        df["startTime"] = pd.to_datetime(df["startTime"])

        return df

    @functools.cached_property
    def calorie_log(self) -> dask.dataframe.DataFrame:
        return (
            dask.dataframe.read_json(
                self.root / "Physical Activity" / "calories-*.json",
                orient="records",
                lines=False,
            )
            .rename(columns={"values": "totalCalories"})
            .set_index("dateTime")
        )

    @functools.cached_property
    def resting_heart_rate_log(self) -> dask.dataframe.DataFrame:
        return (
            dask.dataframe.read_json(
                self.root / "Physical Activity" / "resting_heart_rate-*.json",
                orient="records",
                lines=False,
            )
            .rename(columns={"values": "restingHeartRate"})
            .set_index("dateTime")
        )

    # Level 1 data
    # The "raw data" detected by the FitBit tracker;
    # heart rate (aggregated over 30s), steps (aggregated over 60s),
    # and heart rate variability (aggregated over 5m)

    @functools.cached_property
    def heart_rate_log(self) -> dask.dataframe.DataFrame:
        return dask.dataframe.read_csv(
            self.root / "Physical Activity" / "heart_rate.csv",
            parse_dates=["datetime"],
            blocksize="8MB",
        )

    @functools.cached_property
    def heart_rate_variability_log(self) -> dask.dataframe.DataFrame:
        return dask.dataframe.read_csv(
            self.root / "Sleep" / "Heart Rate Variability Details-*.csv",
            parse_dates=["timestamp"],
            blocksize="8MB",
        )

    @functools.cached_property
    def step_log(self) -> dask.dataframe.DataFrame:
        return (
            dask.dataframe.read_json(
                self.root / "Physical Activity" / "steps-*.json",
                orient="records",
                lines=False,
            )
            .rename(columns={"values": "totalSteps"})
            .set_index("dateTime")
        )


DATA = ExportedData.from_path("data")
